@@ -17,8 +17,6 @@ class ManifestTableSegment:
 class ManifestTable:
     columns: typing.List[str]
     """Columns"""
-    id: str
-    """ID"""
     name: str
     """Name"""
     schema: str
@@ -30,7 +28,7 @@ class ManifestTable:
 @dataclasses_json.dataclass_json(letter_case=dataclasses_json.LetterCase.CAMEL)
 @dataclasses.dataclass()
 class Manifest:
-    tables: typing.List[ManifestTable]
+    tables: typing.Dict[str, ManifestTable]
 
 
 MANIFEST_JSON_FORMAT = package_json_format("slice_db.formats", "manifest.json")
